import sys
from logging import DEBUG, basicConfig, getLogger
from pathlib import Path
from types import SimpleNamespace
from typing import Annotated

from rich.console import Console
from rich.logging import RichHandler
from typer import Context, Exit, Option, Typer

from ._version import __version__
from .func import console, output_stubs
from .utils import _get_default_stubs_path

__all__ = ["__version__", "app"]


log = getLogger(__name__)

app = Typer(
    invoke_without_command=True,
    help="[bold]vs-stubs command line interface[/bold]",
    rich_markup_mode="rich",
    pretty_exceptions_enable=False,
    add_completion=False,
)


def _show_version(value: bool) -> None:
    """Show version info and exit"""

    if value:
        console.print(f"{__version__}")
        raise Exit()


input_opt = Option(
    "--input",
    "-i",
    "-I",
    help="Path to the input .pyi file. Use '-' for piping.",
    rich_help_panel="I/O options",
    allow_dash=True,
    path_type=str,
)
output_opt = Option(
    "--output",
    "-o",
    "-O",
    help="Path to write the output .pyi file. '@' to overwrite the input file. '-' for piping.",
    show_default="vapoursynth-stubs/__init__.pyi inside the site-package folder",
    rich_help_panel="I/O options",
    allow_dash=True,
    path_type=str,
)
load_opt = Option(
    "--load",
    "-L",
    help="Load plugins from a folder or a single library file.",
    rich_help_panel="I/O options",
)
template_opt = Option(
    "--template",
    "-T",
    help="Export blank template; excludes existing plugins unless --load or --add is used.",
)
check_opt = Option(
    "--check",
    "-C",
    help="Check for new plugins or new plugin signatures.",
)
update_opt = Option(
    "--update",
    "-U",
    help="Update the current stubs from the input.",
)
quiet_opt = Option(
    "--quiet",
    help="Suppress message output.",
    rich_help_panel="Informations",
)
debug_opt = Option(
    "--debug",
    hidden=True,
)
version_opt = Option(
    "--version",
    "-V",
    callback=_show_version,
    is_eager=True,
    help="Show version info and exit.",
    rich_help_panel="Informations",
)


@app.command(help="Add or update the specified plugins in the stubs")
def add(plugins: list[str], ctx: Annotated[Context, Option(None)]) -> None:
    console.print(f"Adding plugins: {', '.join(plugins)}")

    output_stubs(
        ctx.obj.input_file,
        ctx.obj.output,
        ctx.obj.template,
        ctx.obj.load,
        False,
        False,
        set(plugins),
        None,
    )
    raise Exit()


@app.command(help="Remove the specified plugins from the stubs")
def remove(plugins: list[str], ctx: Annotated[Context, Option(None)]) -> None:
    console.print(f"Removing plugins: {', '.join(plugins)}")

    output_stubs(
        ctx.obj.input_file,
        ctx.obj.output,
        ctx.obj.template,
        ctx.obj.load,
        False,
        False,
        None,
        set(plugins),
    )
    raise Exit()


@app.callback()
def cli_main(
    ctx: Context,
    input: Annotated[str | None, input_opt] = None,
    output: Annotated[str | None, output_opt] = None,
    template: Annotated[bool, template_opt] = False,
    load: Annotated[list[Path] | None, load_opt] = None,
    check: Annotated[bool, check_opt] = False,
    update: Annotated[bool, update_opt] = False,
    quiet: Annotated[bool, quiet_opt] = False,
    debug: Annotated[bool, debug_opt] = False,
    version: Annotated[bool, version_opt] = False,
) -> None:
    """
    Generate or modify VapourSynth stubs
    """

    if quiet:
        console.quiet = True

    if debug:
        basicConfig(level=DEBUG, handlers=[RichHandler(level=DEBUG, console=Console(stderr=True))])

    if version:
        raise Exit()

    if (check or update) and input is None:
        input = str(_get_default_stubs_path())
    else:
        console.print("Running stub generation...")

    input_file = sys.stdin if input == "-" else input

    if output == "@":
        if input_file is None:
            console.print("[red]Error: You must provide an input_file when output is '@'.[/red]")
            raise Exit(1)
        output_file = input_file
    elif output == "-":
        output_file = sys.stdout
    else:
        output_file = _get_default_stubs_path() if not output else Path(output).with_suffix(".pyi")

    ctx.obj = SimpleNamespace()
    ctx.obj.input_file = input_file
    ctx.obj.output = output_file
    ctx.obj.template = template
    ctx.obj.load = load
    ctx.obj.check = check
    ctx.obj.quiet = quiet

    if ctx.invoked_subcommand is None:
        output_stubs(input_file, output_file, template, load, check, update)
        raise Exit()



_CORE_IMPL_START = "# <plugins/bound/{core_name}>"
_CORE_IMPL_END = "# </plugins/bound/{core_name}>"
_ATTR_IMPL_START = "# <attribute/{core_name}_bound/{name}>"
_ATTR_IMPL_END = "# </attribute/{core_name}_bound/{name}>"

_PLUGINS_IMPL_START = "# <plugins/implementations>"
_PLUGINS_IMPL_END = "# </plugins/implementations>"
_IMPL_START = "# <implementation/{name}>"
_IMPL_END = "# </implementation/{name}>"

_VSCALLBACK_SIGNATURE = "_VSCallback_{plugin}_{func}_{param}"

_callback_signatures = {
    "akarin": {
        "PropExpr": {"dict"},
    },
    "descale": {
        "Decustom": {"custom_kernel"},
        "ScaleCustom": {"custom_kernel"},
    },
    "std": {
        "FrameEval": {"eval"},
        "Lut": {"function"},
        "Lut2": {"function"},
        "ModifyFrame": {"selector"},
    },
    "resize2": {
        "Custom": {"custom_kernel"},
    },
}

_wrappers = {
    "Core": {"FrameEval", "ModifyFrame"},
    "VideoNode": {"FrameEval", "ModifyFrame"},
}



"""Typing stubs for VapourSynth."""

from .func import output_stubs

__version__: str
__version_tuple__: tuple[int | str, ...]

try:
    from ._version import __version__, __version_tuple__
except ImportError:
    __version__ = "0.0.0+unknown"
    __version_tuple__ = (0, 0, 0, "+unknown")


__all__ = ["output_stubs"]



from logging import Handler, LogRecord, StreamHandler
from typing import Any, TextIO

class LogHandle: ...

class PythonVSScriptLoggingBridge(Handler):
    def __init__(self, parent: StreamHandler[TextIO], level: int | str = ...) -> None: ...
    def emit(self, record: LogRecord) -> None: ...

class Error(Exception):
    value: Any
    def __init__(self, value: Any) -> None: ...
    def __str__(self) -> str: ...
    def __repr__(self) -> str: ...



from __future__ import annotations

import sys
from collections.abc import Iterable, Sequence
from functools import cache
from inspect import Parameter
from pathlib import Path
from typing import Any

from vapoursynth import GRAY8, Plugin, core, register_on_destroy

from .constants import _VSCALLBACK_SIGNATURE
from .types import (
    FunctionInterface,
    HasNameSpace,
    PluginInterface,
    UnionLike,
    VSCallbackTypeLike,
    _CoreLike,
    parse_type,
)


def running_via_cli() -> bool:
    # When launched via the installed script (entry point)
    if Path(sys.argv[0]).stem == "vsstubs":
        return True

    # When launched as: python -m vsstubs
    # The main module will be vsstubs.__main__, but argv[0] will be python.
    # So we detect that by checking the *module* that is actually running
    main_mod = sys.modules.get("__main__")
    return bool(main_mod and main_mod.__package__ == "vsstubs")


def _get_typed_dict_repr(d: type) -> str:
    """Workaround function to deal with a badly repr for typed dict."""

    def _pretty_type(value: Any) -> str:
        if isinstance(value, type):
            return str(value.__name__)
        return parse_type(value, True)

    sig = "{" + ", ".join(f'"{n}": {_pretty_type(v)}' for n, v in d.__annotations__.items()) + "}"
    name = d.__name__

    return f'{name} = TypedDict("{name}", {sig})'


def _replace_known_callback_signature(
    param: Parameter, interface: PluginInterface, function: FunctionInterface
) -> Parameter:
    ann = VSCallbackTypeLike(
        _VSCALLBACK_SIGNATURE.format(plugin=interface.namespace, func=function.name, param=param.name)
    )

    if not isinstance(param.annotation, VSCallbackTypeLike):
        ann = UnionLike[*(ann if isinstance(arg, VSCallbackTypeLike) else arg for arg in param.annotation.__args__)]

    return param.replace(annotation=ann)


def _index_by_namespace[HasNameSpaceT: HasNameSpace](impls: Iterable[HasNameSpaceT]) -> dict[str, HasNameSpaceT]:
    return {impl.namespace: impl for impl in impls}


@cache
def _get_default_stubs_path() -> Path:
    import vapoursynth

    return Path(vapoursynth.__file__).parent / "vapoursynth-stubs" / "__init__.pyi"


@cache
def _get_plugins() -> Sequence[Plugin]:
    return tuple(core.plugins())


@cache
def _get_dir(has_dir: Any) -> list[str]:
    return dir(has_dir)


@cache
def _get_cores() -> Sequence[_CoreLike]:
    return [
        core.core,
        core.std.BlankClip(None, 1, 1, GRAY8, 1, 1, 1, 0, True),
        core.std.BlankAudio(None, length=1, keep=True),
    ]


register_on_destroy(_get_plugins.cache_clear)
register_on_destroy(_get_dir.cache_clear)
register_on_destroy(_get_cores.cache_clear)



from collections.abc import Callable
from ctypes import c_void_p
from typing import Any, Final, Literal, Self, final, overload
from weakref import ReferenceType

from ._typing import _IntLike
from .constants import MessageType

# Environment SubSystem
@final
class EnvironmentData: ...

class EnvironmentPolicy:
    def on_policy_registered(self, special_api: EnvironmentPolicyAPI) -> None: ...
    def on_policy_cleared(self) -> None: ...
    def get_current_environment(self) -> EnvironmentData | None: ...
    def set_environment(self, environment: EnvironmentData | None) -> EnvironmentData | None: ...
    def is_alive(self, environment: EnvironmentData) -> bool: ...

@final
class StandaloneEnvironmentPolicy:
    def on_policy_registered(self, api: EnvironmentPolicyAPI) -> None: ...
    def on_policy_cleared(self) -> None: ...
    def get_current_environment(self) -> EnvironmentData: ...
    def set_environment(self, environment: EnvironmentData | None) -> EnvironmentData: ...
    def is_alive(self, environment: EnvironmentData) -> bool: ...
    def _on_log_message(self, level: MessageType, msg: str) -> None: ...

@final
class VSScriptEnvironmentPolicy:
    def on_policy_registered(self, policy_api: EnvironmentPolicyAPI) -> None: ...
    def on_policy_cleared(self) -> None: ...
    def get_current_environment(self) -> EnvironmentData | None: ...
    def set_environment(self, environment: EnvironmentData | None) -> EnvironmentData | None: ...
    def is_alive(self, environment: EnvironmentData) -> bool: ...

@final
class EnvironmentPolicyAPI:
    def wrap_environment(self, environment_data: EnvironmentData) -> Environment: ...
    def create_environment(self, flags: _IntLike = 0) -> EnvironmentData: ...
    def set_logger(self, env: EnvironmentData, logger: Callable[[int, str], None]) -> None: ...
    def get_vapoursynth_api(self, version: int) -> c_void_p: ...
    def get_core_ptr(self, environment_data: EnvironmentData) -> c_void_p: ...
    def destroy_environment(self, env: EnvironmentData) -> None: ...
    def unregister_policy(self) -> None: ...

def register_policy(policy: EnvironmentPolicy) -> None: ...
def has_policy() -> bool: ...
def register_on_destroy(callback: Callable[..., None]) -> None: ...
def unregister_on_destroy(callback: Callable[..., None]) -> None: ...
def _try_enable_introspection(version: int | None = None) -> bool: ...
@final
class _FastManager:
    def __enter__(self) -> None: ...
    def __exit__(self, *_: object) -> None: ...

class Environment:
    env: Final[ReferenceType[EnvironmentData]]
    def __repr__(self) -> str: ...
    @overload
    def __eq__(self, other: Environment) -> bool: ...
    @overload
    def __eq__(self, other: object) -> bool: ...
    @property
    def alive(self) -> bool: ...
    @property
    def single(self) -> bool: ...
    @classmethod
    def is_single(cls) -> bool: ...
    @property
    def env_id(self) -> int: ...
    @property
    def active(self) -> bool: ...
    def copy(self) -> Self: ...
    def use(self) -> _FastManager: ...

class Local:
    def __getattr__(self, key: str) -> Any: ...
    def __setattr__(self, key: str, value: Any) -> None: ...
    def __delattr__(self, key: str) -> None: ...

def get_current_environment() -> Environment: ...

class CoreTimings:
    def __repr__(self) -> str: ...
    def __str__(self) -> str: ...
    @property
    def enabled(self) -> bool: ...
    @enabled.setter
    def enabled(self, enabled: bool) -> bool: ...
    @property
    def freed_nodes(self) -> bool: ...
    @freed_nodes.setter
    def freed_nodes(self, value: Literal[0]) -> bool: ...



from collections.abc import Callable
from typing import Concatenate, overload

from ._typing import (
    _SequenceLike,
    _VSCallback_std_FrameEval_eval,
    _VSCallback_std_FrameEval_eval_0,
    _VSCallback_std_FrameEval_eval_1,
    _VSCallback_std_FrameEval_eval_2,
    _VSCallback_std_FrameEval_eval_3,
    _VSCallback_std_ModifyFrame_selector,
    _VSCallback_std_ModifyFrame_selector_0,
    _VSCallback_std_ModifyFrame_selector_1,
)
from .nodes import VideoNode
from .plugin import Function, Plugin

_VSPlugin = Plugin
_VSFunction = Function

class _Wrapper:
    class Function[**P, R](_VSFunction):
        def __init__[PluginT: Plugin](self, function: Callable[Concatenate[PluginT, P], R]) -> None: ...
        def __call__(self, *args: P.args, **kwargs: P.kwargs) -> R: ...

class _Wrapper_Core_bound_FrameEval:
    class Function(_VSFunction):
        def __init__[PluginT: Plugin](self, function: Callable[Concatenate[PluginT, ...], VideoNode]) -> None: ...
        @overload
        def __call__(
            self,
            clip: VideoNode,
            eval: _VSCallback_std_FrameEval_eval_0,
            prop_src: None = None,
            clip_src: VideoNode | _SequenceLike[VideoNode] | None = None,
        ) -> VideoNode: ...
        @overload
        def __call__(
            self,
            clip: VideoNode,
            eval: _VSCallback_std_FrameEval_eval_1,
            prop_src: VideoNode,
            clip_src: VideoNode | _SequenceLike[VideoNode] | None = None,
        ) -> VideoNode: ...
        @overload
        def __call__(
            self,
            clip: VideoNode,
            eval: _VSCallback_std_FrameEval_eval_2,
            prop_src: _SequenceLike[VideoNode],
            clip_src: VideoNode | _SequenceLike[VideoNode] | None = None,
        ) -> VideoNode: ...
        @overload
        def __call__(
            self,
            clip: VideoNode,
            eval: _VSCallback_std_FrameEval_eval_3,
            prop_src: VideoNode | _SequenceLike[VideoNode],
            clip_src: VideoNode | _SequenceLike[VideoNode] | None = None,
        ) -> VideoNode: ...
        @overload
        def __call__(
            self,
            clip: VideoNode,
            eval: _VSCallback_std_FrameEval_eval,
            prop_src: VideoNode | _SequenceLike[VideoNode] | None,
            clip_src: VideoNode | _SequenceLike[VideoNode] | None = None,
        ) -> VideoNode: ...

class _Wrapper_VideoNode_bound_FrameEval:
    class Function(_VSFunction):
        def __init__[PluginT: Plugin](self, function: Callable[Concatenate[PluginT, ...], VideoNode]) -> None: ...
        @overload
        def __call__(
            self,
            eval: _VSCallback_std_FrameEval_eval_0,
            prop_src: None = None,
            clip_src: VideoNode | _SequenceLike[VideoNode] | None = None,
        ) -> VideoNode: ...
        @overload
        def __call__(
            self,
            eval: _VSCallback_std_FrameEval_eval_1,
            prop_src: VideoNode,
            clip_src: VideoNode | _SequenceLike[VideoNode] | None = None,
        ) -> VideoNode: ...
        @overload
        def __call__(
            self,
            eval: _VSCallback_std_FrameEval_eval_2,
            prop_src: _SequenceLike[VideoNode],
            clip_src: VideoNode | _SequenceLike[VideoNode] | None = None,
        ) -> VideoNode: ...
        @overload
        def __call__(
            self,
            eval: _VSCallback_std_FrameEval_eval_3,
            prop_src: VideoNode | _SequenceLike[VideoNode],
            clip_src: VideoNode | _SequenceLike[VideoNode] | None = None,
        ) -> VideoNode: ...
        @overload
        def __call__(
            self,
            eval: _VSCallback_std_FrameEval_eval,
            prop_src: VideoNode | _SequenceLike[VideoNode] | None,
            clip_src: VideoNode | _SequenceLike[VideoNode] | None = None,
        ) -> VideoNode: ...

class _Wrapper_Core_bound_ModifyFrame:
    class Function(_VSFunction):
        def __init__[PluginT: Plugin](self, function: Callable[Concatenate[PluginT, ...], VideoNode]) -> None: ...
        @overload
        def __call__(
            self, clip: VideoNode, clips: VideoNode, selector: _VSCallback_std_ModifyFrame_selector_0
        ) -> VideoNode: ...
        @overload
        def __call__(
            self, clip: VideoNode, clips: _SequenceLike[VideoNode], selector: _VSCallback_std_ModifyFrame_selector_1
        ) -> VideoNode: ...
        @overload
        def __call__(
            self,
            clip: VideoNode,
            clips: VideoNode | _SequenceLike[VideoNode],
            selector: _VSCallback_std_ModifyFrame_selector,
        ) -> VideoNode: ...

class _Wrapper_VideoNode_bound_ModifyFrame:
    class Function(_VSFunction):
        def __init__[PluginT: Plugin](self, function: Callable[Concatenate[PluginT, ...], VideoNode]) -> None: ...
        @overload
        def __call__(self, clips: VideoNode, selector: _VSCallback_std_ModifyFrame_selector_0) -> VideoNode: ...
        @overload
        def __call__(
            self, clips: _SequenceLike[VideoNode], selector: _VSCallback_std_ModifyFrame_selector_1
        ) -> VideoNode: ...
        @overload
        def __call__(
            self, clips: VideoNode | _SequenceLike[VideoNode], selector: _VSCallback_std_ModifyFrame_selector
        ) -> VideoNode: ...



from collections.abc import Sequence
from logging import getLogger
from os import PathLike
from pathlib import Path
from typing import IO, Any

from rich.console import Console

from .stubs import (
    construct_implementation,
    get_implementations_from_input,
    load_plugins,
    retrieve_plugins,
    write_implementations,
    write_plugins_bound,
)
from .template import get_template
from .types import Implementation, parse_type
from .utils import _get_cores, _index_by_namespace, running_via_cli

log, console = getLogger(__name__), Console(stderr=True)


def output_stubs(
    input_file: str | PathLike[str] | IO[str] | None,
    output: str | PathLike[str] | IO[str],
    template: bool = False,
    load: Sequence[str | PathLike[str]] | None = None,
    check: bool = False,
    update: bool = False,
    add: set[str] | None = None,
    remove: set[str] | None = None,
) -> None:
    """
    Generate or update VapourSynth stub files.

    This function creates a `.pyi` stub file based on an existing stub, a blank template,
    or additional plugin definitions.
    It can also validate stubs against newly detected plugins or signatures.

    Args:
        input_file: Optional path to an existing `.pyi` file to use as the base for generating stubs.
            If None, a new stub is created from scratch.

        output: Path to the `.pyi` file where the generated stubs will be written.

        template: If True, generate a blank template with no existing plugins
            unless explicitly provided via `load` or `add`.

        load: One or more paths to plugin definitions (either directories or individual library files)
            to be included in the stubs.

        check: If True, validate the generated stubs against newly discovered plugins or signatures,
            reporting any discrepancies.

        update: If True, only update the current stubs from the input_file.

        add: A set of plugin names to add or update in the stubs.

        remove: A set of plugin names to remove from the stubs.
    """

    if not running_via_cli():
        console.quiet = True

    if load:
        console.print(f"Loading plugins from: {load}")
        plugins_to_add = load_plugins(load)
        add = plugins_to_add if not add else add | plugins_to_add

    cores = _get_cores()
    pinters = retrieve_plugins(cores)

    if input_file:
        tmpl = Path(input_file).read_text() if isinstance(input_file, (str, PathLike)) else input_file.read()

        implementations = get_implementations_from_input(tmpl)

        if check:
            console.print("Checking stubs...")

            old_impl = _index_by_namespace(implementations)
            new_impl = _index_by_namespace(construct_implementation(pinter) for pinter in pinters)

            old_keys, new_keys = set(old_impl), set(new_impl)

            only_old = old_keys - new_keys
            only_new = new_keys - old_keys

            if only_old or only_new:
                console.print(
                    f"[yellow]"
                    f"Mismatched plugin(s): "
                    f"only in input={', '.join(sorted(only_old)) or 'none'}, "
                    f"only new={', '.join(sorted(only_new)) or 'none'}"
                    "[/yellow]"
                )

            for ns in old_keys & new_keys:
                _compare_plugins(old_impl[ns], new_impl[ns], ns)
        elif update:
            impl_ns = [i.namespace for i in implementations]

            console.print(f"Updating stubs... Found {len(impl_ns)} plugins to update: {impl_ns}")

            implementations = [construct_implementation(pinter) for pinter in pinters if pinter.namespace in impl_ns]

    elif template:
        tmpl = get_template()
        implementations = []
    else:
        tmpl = get_template()
        implementations = [construct_implementation(pinter) for pinter in pinters]

    if add or remove:
        impl_map = _index_by_namespace(implementations)

        log.debug("add: %s", add)
        log.debug("remove: %s", remove)

        warn_msg = '[yellow]"{ns}" isn\'t a valid plugin namespace.[/yellow]'

        if add:
            pinters_map = _index_by_namespace(pinters)

            for ns in add:
                if ns not in pinters_map:
                    console.print(warn_msg.format(ns=ns))
                    continue

                impl_map[ns] = construct_implementation(pinters_map[ns])

        if remove:
            for ns in remove:
                if ns not in impl_map:
                    console.print(warn_msg.format(ns=ns))
                    continue

                del impl_map[ns]

        implementations = list(impl_map.values())

    log.debug("parse_type: %s", parse_type.cache_info())

    tmpl = write_implementations(implementations, tmpl)
    tmpl = write_plugins_bound(implementations, tmpl)

    log.debug("output: %r", output)

    if isinstance(output, (str, PathLike)):
        output = Path(output)
        output.parent.mkdir(parents=True, exist_ok=True)
        output.write_text(tmpl)
    else:
        output.write(tmpl)

    console.print("[green]Done![/green]")


def _compare_plugins(old: Implementation, new: Implementation, ns: str) -> None:
    checks: list[tuple[str, Any, Any]] = [
        ("functions", dict(old.functions), dict(new.functions)),
        ("description", old.description, new.description),
        ("extra types", old.extra_types, new.extra_types),
    ]
    for field, old_val, new_val in checks:
        if old_val != new_val:
            console.print(f'For the plugin {ns}, the "{field}" differ.')



from collections.abc import Iterator
from ctypes import c_void_p
from types import TracebackType
from typing import Final, Literal, Self, SupportsIndex

from ._typing import _IntLike, _SupportsKeysAndGetItem, _VSValue
from .constants import SampleType
from .formats import ChannelLayout, VideoFormat
from .props import FrameProps

class FramePtr: ...

# These memoryview-likes don't exist at runtime.
class _video_view(memoryview):  # type: ignore[misc]
    def __getitem__(self, index: tuple[int, int]) -> float: ...  # type: ignore[override]
    def __setitem__(self, index: tuple[int, int], other: float) -> None: ...  # type: ignore[override]
    @property
    def shape(self) -> tuple[int, int]: ...
    @property
    def strides(self) -> tuple[int, int]: ...
    @property
    def ndim(self) -> Literal[2]: ...
    @property
    def obj(self) -> FramePtr: ...  # type: ignore[override]
    def tolist(self) -> list[float]: ...  # type: ignore[override]

class _audio_view(memoryview):  # type: ignore[misc]
    def __getitem__(self, index: int) -> float: ...  # type: ignore[override]
    def __setitem__(self, index: int, other: float) -> None: ...  # type: ignore[override]
    @property
    def shape(self) -> tuple[int]: ...
    @property
    def strides(self) -> tuple[int]: ...
    @property
    def ndim(self) -> Literal[1]: ...
    @property
    def obj(self) -> FramePtr: ...  # type: ignore[override]
    def tolist(self) -> list[float]: ...  # type: ignore[override]

class RawFrame:
    def __repr__(self) -> str: ...
    def __str__(self) -> str: ...
    def __enter__(self) -> Self: ...
    def __exit__(
        self, exc: type[BaseException] | None = None, val: BaseException | None = None, tb: TracebackType | None = None
    ) -> bool | None: ...
    def __getitem__(self, index: SupportsIndex) -> memoryview: ...
    def __len__(self) -> int: ...
    @property
    def closed(self) -> bool: ...
    @property
    def props(self) -> FrameProps: ...
    @props.setter
    def props(self, new_props: _SupportsKeysAndGetItem[str, _VSValue]) -> None: ...
    @property
    def readonly(self) -> bool: ...
    def copy(self) -> Self: ...
    def close(self) -> None: ...
    def get_write_ptr(self, plane: _IntLike) -> c_void_p: ...
    def get_read_ptr(self, plane: _IntLike) -> c_void_p: ...
    def get_stride(self, plane: _IntLike) -> int: ...

# Behave like a Sequence
class VideoFrame(RawFrame):
    format: Final[VideoFormat]
    width: Final[int]
    height: Final[int]

    def __getitem__(self, index: SupportsIndex) -> _video_view: ...
    def readchunks(self) -> Iterator[_video_view]: ...

# Behave like a Sequence
class AudioFrame(RawFrame):
    sample_type: Final[SampleType]
    bits_per_sample: Final[int]
    bytes_per_sample: Final[int]
    channel_layout: Final[int]
    num_channels: Final[int]

    def __getitem__(self, index: SupportsIndex) -> _audio_view: ...
    @property
    def channels(self) -> ChannelLayout: ...



from __future__ import annotations

import re
from collections import defaultdict
from collections.abc import Iterable, Mapping, Sequence
from functools import cache
from inspect import Parameter
from itertools import takewhile
from logging import getLogger
from os import PathLike
from pathlib import Path
from typing import Any, is_typeddict

from vapoursynth import Error, core

from .constants import (
    _ATTR_IMPL_END,
    _ATTR_IMPL_START,
    _CORE_IMPL_END,
    _CORE_IMPL_START,
    _IMPL_END,
    _IMPL_START,
    _PLUGINS_IMPL_END,
    _PLUGINS_IMPL_START,
    _callback_signatures,
    _wrappers,
)
from .types import (
    FunctionInterface,
    Implementation,
    PluginInterface,
    VideoNodeType,
    WrappedFunction,
    _CoreLike,
    parse_type,
)
from .utils import _get_cores, _get_dir, _get_plugins, _get_typed_dict_repr, _replace_known_callback_signature

log = getLogger(__name__)


def load_plugins(paths: Iterable[str | PathLike[str]]) -> set[str]:
    """
    Load the plugins from a list of dll or path folders.

    Returns the new loaded namespace plugins.
    """
    old_plugins = {p.namespace for p in core.plugins()}

    for path in paths:
        path = Path(path)

        if not path.exists():
            raise ValueError(f'This path "{path}" doesn\'t exist.')

        if path.is_dir():
            core.std.LoadAllPlugins(str(path))
        else:
            # std.LoadAllPlugins silently skips the plugins that fail to load.
            try:
                core.std.LoadPlugin(str(path))
            except Error as e:
                log.exception(e)

    return {p.namespace for p in _get_plugins()} - old_plugins


def retrieve_plugins(core_like: Sequence[_CoreLike]) -> Sequence[PluginInterface]:
    """
    Get a sequence of PluginInferface.

    A PluginInferface offers an interface for each core-like it owns with its functions attached to it.
    """
    plugins = list[PluginInterface]()

    for plugin in _get_plugins():
        functions = defaultdict[str, list[FunctionInterface]](list)

        for cl in core_like:
            # Some plugins only have vs.Core as bound core
            if plugin.namespace not in _get_dir(cl):
                continue

            # Get the actual plugin attached to its core to get the right functions signatures.
            if cl is not core.core:
                plugin = getattr(cl, plugin.namespace)

            functions[cl.__class__.__name__].extend(
                # Only gets the functions that __dir__ provides
                FunctionInterface(f.name, f.__signature__)
                for f in plugin.functions()
                if f.name in _get_dir(plugin)
            )

        plugins.append(PluginInterface(plugin.namespace, functions, plugin.name))

    return plugins


def construct_implementation(interface: PluginInterface) -> Implementation:
    """Contructs a full implementation block with all the functions for all the cores-like."""

    functions_map = dict[str, list[WrappedFunction]]()
    extras = list[str]()

    for core_name, functions in interface.functions.items():
        functions_list = list[WrappedFunction]()

        for function in functions:
            parameters = function.signature.parameters.copy()

            for k, v in parameters.items():
                parameters[k] = v.replace(annotation=parse_type(v.annotation))

            # Replaces the anonymous callback types to known signatures.
            if param_names := _callback_signatures.get(interface.namespace, {}).get(function.name):
                for name in param_names:
                    parameters[name] = _replace_known_callback_signature(parameters[name], interface, function)

            # If a function returns Any, it's probably a APIv3 plugin.
            # We assume it always returns a VideoNode but we know it's not always the case...
            if function.signature.return_annotation == Any:
                return_annotation = VideoNodeType()
                log.debug("APIv3 plugin detected: '%s.%s.%s'", core_name, interface.namespace, function.name)
            else:
                return_annotation = parse_type(function.signature.return_annotation, True)

            signature = function.signature.replace(
                parameters=(Parameter("self", Parameter.POSITIONAL_OR_KEYWORD), *parameters.values()),
                return_annotation=return_annotation,
            )

            wrapper = (
                f"_Wrapper_{core_name}_bound_{function.name}.Function"
                if function.name in _wrappers.get(core_name, set())
                else None
            )

            functions_list.append(WrappedFunction(f"{function.name}{signature}: ...", wrapper))

            if is_typeddict(td := signature.return_annotation):
                extras.append(_get_typed_dict_repr(td))

        functions_map[core_name] = functions_list

    return Implementation(interface.namespace, functions_map, interface.description, extras)


def get_implementations_from_input(text: str) -> list[Implementation]:
    """Parse a file to extract plugin implementations."""

    plugins_impl_block_regex = rf"{_PLUGINS_IMPL_START}(.*?){_PLUGINS_IMPL_END}"

    plugins_impl_block_matched = re.search(plugins_impl_block_regex, text, re.DOTALL)

    if not plugins_impl_block_matched:
        return []

    # Regex to capture class blocks like "_Core_bound", "_VideoNode_bound" or "_AudioNode_bound"
    plugins_impl_block_pattern_text = r"class _(\w+)_bound:\s*class Plugin.*?:([\s\S]*?)(?=\n\s*class _\w+_bound:|\Z)"
    plugins_impl_block_pattern = re.compile(plugins_impl_block_pattern_text, re.MULTILINE)

    # Regex to capture decorator + full function definition lines
    func_pattern_text = r"^\s*(?:@([^\n]*))?\s*\n?\s*def ([\s\S]+?-> [\w\[\], |]+: \.\.\.)"
    func_pattern = re.compile(func_pattern_text, re.MULTILINE)

    implementations = list[Implementation]()

    # Regex to capture <implementation/{plugin_name}>
    impl_capture_regex = rf"{_IMPL_START.format(name='([^>]+)')}(.*?){_IMPL_END.format(name='\\1')}"

    for name, body in re.findall(impl_capture_regex, plugins_impl_block_matched.group(), re.DOTALL):
        if not body:
            raise ValueError(f"No plugin implementation block found for {name}.")

        extras = list(takewhile(lambda s: not s.startswith("class"), (s for s in body.splitlines() if s)))

        functions = defaultdict[str, list[WrappedFunction]](list)

        for core_like, fbody in plugins_impl_block_pattern.findall(body):
            for decorator, func_def in func_pattern.findall(fbody):
                # Normalize the function definition into one stripped line
                normalized_func = " ".join(func_def.split()).replace("( ", "(").replace(", )", ")")

                # Store functions under the core_like key
                functions[core_like].append(WrappedFunction(normalized_func, decorator))

        doc = _extract_description(text, name, functions)

        implementations.append(Implementation(name, functions, doc, extras))

    return implementations


@cache
def _search_core_impl(core_name: str, text: str) -> re.Match[str] | None:
    """Regex to capture the whole plugins bound block"""
    return re.search(rf"{_CORE_IMPL_START}(.*?){_CORE_IMPL_END}".format(core_name=core_name), text, re.DOTALL)


def _extract_description(text: str, ns: str, functions: Mapping[str, Sequence[WrappedFunction]]) -> str:
    core_name = next(iter(functions))

    core_impl_matched = _search_core_impl(core_name, text)

    if core_impl_matched is None:
        raise ValueError(f"No core implementation block found for {core_name}.")

    # Regex to capture the attribute block
    attr_matched = re.search(
        rf"{_ATTR_IMPL_START}(.*?){_ATTR_IMPL_END}".format(core_name=core_name, name=ns),
        core_impl_matched.group(0),
        re.DOTALL,
    )
    if not attr_matched:
        raise ValueError(f"No attribute block found for {core_name}.")

    doc_matched = re.search(r"(?:\"\"\"(.*?)\"\"\"|)$", attr_matched.group(1), re.DOTALL)

    return "" if not doc_matched else doc_matched.group(1)


def write_implementations(implementations: list[Implementation], template: str) -> str:
    """Replace the plugin implementations block in `template` with the given implementations."""
    body = "\n".join(impl.as_stub() for impl in sorted(implementations))
    replacement = f"{_PLUGINS_IMPL_START}\n{body}\n{_PLUGINS_IMPL_END}"

    return re.sub(
        rf"{_PLUGINS_IMPL_START}.*{_PLUGINS_IMPL_END}",
        replacement,
        template,
        flags=re.DOTALL,
    )


def write_plugins_bound(implementations: list[Implementation], template: str) -> str:
    """Replace the plugin bound blocks in `template` with the given implementations."""
    for core_like in _get_cores():
        cname = core_like.__class__.__name__

        plugins_bound = "\n".join(
            (
                f"{_ATTR_IMPL_START.format(core_name=cname, name=i.namespace)}\n"
                f"    {i.namespace}: Final[_{i.namespace}._{cname}_bound.Plugin]\n"
                f'    """{i.description}"""\n'
                f"{_ATTR_IMPL_END.format(core_name=cname, name=i.namespace)}"
            )
            for i in sorted(implementations, key=lambda i: i.namespace)
            if cname in i.functions
        )

        template = re.sub(
            rf"{_CORE_IMPL_START}.*{_CORE_IMPL_END}".format(core_name=cname),
            rf"{_CORE_IMPL_START}\n{plugins_bound}\n{_CORE_IMPL_END}".format(core_name=cname),
            template,
            flags=re.DOTALL,
        )

    return template



from types import MappingProxyType
from typing import Literal, NamedTuple

from ._typing import _IntLike
from .nodes import AudioNode, VideoNode

class VideoOutputTuple(NamedTuple):
    clip: VideoNode
    alpha: VideoNode | None
    alt_output: Literal[0, 1, 2]

def clear_output(index: _IntLike = 0) -> None: ...
def clear_outputs() -> None: ...
def get_outputs() -> MappingProxyType[int, VideoOutputTuple | AudioNode]: ...
def get_output(index: _IntLike = 0) -> VideoOutputTuple | AudioNode: ...



from collections.abc import Callable, Iterator
from concurrent.futures import Future
from fractions import Fraction
from typing import IO, Any, Final, Literal, Self, overload
from warnings import deprecated

from ._typing import _IntLike, _VSValue
from .constants import FilterMode, MessageType, SampleType, VapourSynthAPIVersion, VapourSynthVersion
from .env import CoreTimings
from .formats import ChannelLayout, VideoFormat
from .frames import AudioFrame, RawFrame, VideoFrame
from .logging import LogHandle
from .plugin import Plugin

class RawNode:
    def __repr__(self) -> str: ...
    def __str__(self) -> str: ...
    def __eq__(self, other: object) -> bool: ...
    def __hash__(self) -> int: ...
    def __dir__(self) -> list[str]: ...
    def __getitem__(self, index: int | slice[int | None, int | None, int | None]) -> Self: ...
    def __len__(self) -> int: ...
    def __add__(self, other: Self) -> Self: ...
    def __mul__(self, other: int) -> Self: ...
    def __getattr__(self, name: str) -> Plugin: ...
    @property
    def node_name(self) -> str: ...
    @property
    def timings(self) -> int: ...
    @timings.setter
    def timings(self, value: Literal[0]) -> None: ...
    @property
    def mode(self) -> FilterMode: ...
    @property
    def dependencies(self) -> tuple[Self, ...]: ...
    @property
    def _name(self) -> str: ...
    @property
    def _inputs(self) -> dict[str, _VSValue]: ...
    def get_frame(self, n: _IntLike) -> RawFrame: ...
    @overload
    def get_frame_async(self, n: _IntLike) -> Future[RawFrame]: ...
    @overload
    def get_frame_async(self, n: _IntLike, cb: Callable[[RawFrame | None, Exception | None], None]) -> None: ...
    def frames(
        self, prefetch: int | None = None, backlog: int | None = None, close: bool = False
    ) -> Iterator[RawFrame]: ...
    def set_output(self, index: _IntLike = 0) -> None: ...
    def clear_cache(self) -> None: ...
    def is_inspectable(self, version: int | None = None) -> bool: ...

type _CurrentFrame = int
type _TotalFrames = int

# Behave like a Sequence
class VideoNode(RawNode):
    format: Final[VideoFormat]
    width: Final[int]
    height: Final[int]
    num_frames: Final[int]
    fps_num: Final[int]
    fps_den: Final[int]
    fps: Final[Fraction]
    def get_frame(self, n: _IntLike) -> VideoFrame: ...
    @overload  # type: ignore[override]
    def get_frame_async(self, n: _IntLike) -> Future[VideoFrame]: ...
    @overload
    def get_frame_async(  # pyright: ignore[reportIncompatibleMethodOverride]
        self, n: _IntLike, cb: Callable[[VideoFrame | None, Exception | None], None]
    ) -> None: ...
    def frames(
        self, prefetch: int | None = None, backlog: int | None = None, close: bool = False
    ) -> Iterator[VideoFrame]: ...
    def set_output(self, index: _IntLike = 0, alpha: Self | None = None, alt_output: Literal[0, 1, 2] = 0) -> None: ...
    def output(
        self,
        fileobj: IO[bytes],
        y4m: bool = False,
        progress_update: Callable[[_CurrentFrame, _TotalFrames], None] | None = None,
        prefetch: int = 0,
        backlog: int = -1,
    ) -> None: ...

# <plugins/bound/VideoNode>
# </plugins/bound/VideoNode>

# Behave like a Sequence
class AudioNode(RawNode):
    sample_type: Final[SampleType]
    bits_per_sample: Final[int]
    bytes_per_sample: Final[int]
    channel_layout: Final[int]
    num_channels: Final[int]
    sample_rate: Final[int]
    num_samples: Final[int]
    num_frames: Final[int]
    @property
    def channels(self) -> ChannelLayout: ...
    def get_frame(self, n: _IntLike) -> AudioFrame: ...
    @overload  # type: ignore[override]
    def get_frame_async(self, n: _IntLike) -> Future[AudioFrame]: ...
    @overload
    def get_frame_async(  # pyright: ignore[reportIncompatibleMethodOverride]
        self, n: _IntLike, cb: Callable[[AudioFrame | None, Exception | None], None]
    ) -> None: ...
    def frames(
        self, prefetch: int | None = None, backlog: int | None = None, close: bool = False
    ) -> Iterator[AudioFrame]: ...

# <plugins/bound/AudioNode>
# </plugins/bound/AudioNode>

class Core:
    timings: Final[CoreTimings]
    def __repr__(self) -> str: ...
    def __str__(self) -> str: ...
    def __dir__(self) -> list[str]: ...
    def __getattr__(self, name: str) -> Plugin: ...
    @property
    def api_version(self) -> VapourSynthAPIVersion: ...
    @property
    def core_version(self) -> VapourSynthVersion: ...
    @property
    def num_threads(self) -> int: ...
    @num_threads.setter
    def num_threads(self, value: _IntLike) -> None: ...
    @property
    def max_cache_size(self) -> int: ...
    @max_cache_size.setter
    def max_cache_size(self, mb: _IntLike) -> None: ...
    @property
    def used_cache_size(self) -> int: ...
    @property
    def flags(self) -> int: ...
    def plugins(self) -> Iterator[Plugin]: ...
    def query_video_format(
        self,
        color_family: _IntLike,
        sample_type: _IntLike,
        bits_per_sample: _IntLike,
        subsampling_w: _IntLike = 0,
        subsampling_h: _IntLike = 0,
    ) -> VideoFormat: ...
    def get_video_format(self, id: _IntLike) -> VideoFormat: ...
    def create_video_frame(self, format: VideoFormat, width: _IntLike, height: _IntLike) -> VideoFrame: ...
    def log_message(self, message_type: _IntLike, message: str) -> None: ...
    def add_log_handler(self, handler_func: Callable[[MessageType, str], None]) -> LogHandle: ...
    def remove_log_handler(self, handle: LogHandle) -> None: ...
    def clear_cache(self) -> None: ...
    @deprecated("core.version() is deprecated, use str(core)!", category=DeprecationWarning)
    def version(self) -> str: ...
    @deprecated(
        "core.version_number() is deprecated, use core.core_version.release_major!", category=DeprecationWarning
    )
    def version_number(self) -> int: ...

# <plugins/bound/Core>
# </plugins/bound/Core>

# _CoreProxy doesn't inherit from Core but __getattr__ returns the attribute from the actual core
class _CoreProxy(Core):
    def __setattr__(self, name: str, value: Any) -> None: ...
    @property
    def core(self) -> Core: ...

core: _CoreProxy

# <plugins/implementations>
# </plugins/implementations>



from collections.abc import Iterator
from enum import Enum
from inspect import Signature
from typing import Any

from .formats import VideoFormat
from .plugin import Function

def construct_signature(
    signature: str | Function, return_signature: str, injected: bool | None = None, name: str | None = None
) -> Signature: ...
def _construct_type(signature: str) -> Any: ...
def _construct_parameter(signature: str) -> Any: ...
def _construct_repr_wrap(value: str | Enum | VideoFormat | Iterator[str]) -> str: ...
def _construct_repr(obj: Any, **kwargs: Any) -> str: ...
